@@ -1,32 +1,28 @@
-from __future__ import division
+from __future__ import division  # TODO: refactor
 
 from django.utils.functional import cached_property
-
-try:
-    from collections import OrderedDict
-except ImportError:
-    from ordereddict import OrderedDict
+from django.utils.module_loading import import_string
 
 from decimal import Decimal
 from django_countries.fields import CountryField
 from django_iban.fields import IBANField, SWIFTBICField
 from djmoney.forms.widgets import CURRENCY_CHOICES
-from jsonfield import JSONField  # TODO: use native postgresql JSONfield instead
 from model_utils import Choices
 from model_utils.fields import MonitorField
 
 from django.conf import settings
-from django.core.exceptions import ImproperlyConfigured
-
-try:
-    from django.core.urlresolvers import reverse
-except ImportError:
-    from django.urls import reverse
-
 from django.core.validators import EMPTY_VALUES, MaxValueValidator, MinValueValidator
 from django.db import models, transaction
 from django.db.models import Max
-from django.template import Template, Context
+
+try:
+    # Django 3.1
+    from django.db.models import JSONField
+except ImportError:
+    # older Django
+    from django.contrib.postgres.fields import JSONField
+
+from django.urls import reverse
 from django.utils.timezone import now
 from django.utils.translation import ugettext_lazy as _
 
@@ -247,67 +243,21 @@ class Invoice(models.Model):
         )(self)
 
     @staticmethod
-    def get_next_sequence(type, important_date, number_prefix=None):
+    def get_next_sequence(type, important_date, number_prefix=None, related_invoices=None):
         """
-        Returns next invoice sequence based on ``settings.INVOICING_COUNTER_PERIOD``.
-
-        .. warning::
-
-            This is only used to prepopulate ``sequence`` field on saving new invoice.
-            To get invoice sequence always use ``sequence`` field.
-
-        .. note::
-
-            To get invoice number use ``number`` field.
-
-        :return: string (generated next sequence)
+        Returns next invoice sequence based on ``settings.INVOICING_SEQUENCE_GENERATOR``.
         """
-        with transaction.atomic():
-            Invoice.objects.lock()
-
-            invoice_counter_reset = getattr(settings, 'INVOICING_COUNTER_PERIOD', Invoice.COUNTER_PERIOD.YEARLY)
-
-            if invoice_counter_reset == Invoice.COUNTER_PERIOD.DAILY:
-                relative_invoices = Invoice.objects.filter(date_issue=important_date)
-
-            elif invoice_counter_reset == Invoice.COUNTER_PERIOD.YEARLY:
-                relative_invoices = Invoice.objects.filter(date_issue__year=important_date.year)
-
-            elif invoice_counter_reset == Invoice.COUNTER_PERIOD.MONTHLY:
-                relative_invoices = Invoice.objects.filter(date_issue__year=important_date.year, date_issue__month=important_date.month)
-
-            else:
-                raise ImproperlyConfigured("INVOICING_COUNTER_PERIOD can be set only to these values: DAILY, MONTHLY, YEARLY.")
-
-            invoice_counter_per_type = getattr(settings, 'INVOICING_COUNTER_PER_TYPE', False)
-
-            if invoice_counter_per_type:
-                relative_invoices = relative_invoices.filter(type=type)
-
-            if number_prefix is not None:
-                relative_invoices = relative_invoices.filter(number__startswith=number_prefix)
-
-            start_from = getattr(settings, 'INVOICING_NUMBER_START_FROM', 1)
-            last_sequence = relative_invoices.aggregate(Max('sequence'))['sequence__max'] or start_from - 1
-
-            return last_sequence + 1
+        generator = getattr(settings, 'INVOICING_SEQUENCE_GENERATOR', 'invoicing.helpers.sequence_generator')
+        generator = import_string(generator)
+        return generator(type, important_date, number_prefix, related_invoices)
 
     def _get_number(self, number_format=None):
         """
-        Generates on the fly invoice number from template provided by ``settings.INVOICING_NUMBER_FORMAT``.
-        ``Invoice`` object is provided as ``invoice`` variable to the template, therefore all object fields
-        can be used to generate number format.
-
-        .. warning::
-
-            This is only used to prepopulate ``number`` field on saving new invoice.
-            To get invoice number always use ``number`` field.
-
-        :return: string (generated number)
+        Returns next invoice sequence based on ``settings.INVOICING_NUMBER_FORMATTER``.
         """
-        if not number_format:
-            number_format = getattr(settings, "INVOICING_NUMBER_FORMAT", "{{ invoice.date_tax_point|date:'Y' }}/{{ invoice.sequence }}")
-        return Template(number_format).render(Context({'invoice': self}))
+        formatter = getattr(settings, 'INVOICING_NUMBER_FORMATTER', 'invoicing.helpers.number_formatter')
+        formatter = import_string(formatter)
+        return formatter(self, number_format)
 
     def get_tax_rate(self):
         if self.taxation_policy:
